"""Miscellaneous statistical functions"""

import numpy


def coef_variation(x: numpy.ndarray, axis: int = 0) -> numpy.ndarray:
    """Calculates the coefficientof variation of the rows or columns of a matrix.
    The coefficient of variation is given by the standard deviation divided by the mean of a variable:

    .. math:: \\frac{\\sigma}{\\mu}

    :param x: the data matrix
    :param axis: `0` to calculate for columns, `1` for rows
    :returns: a vector containing the coefficient of variations along the chosen axis
    """
    if not isinstance(x, numpy.matrix):
        x = numpy.asarray(x)

    mean = numpy.mean(x, axis=axis)
    stddev = numpy.std(x, axis=axis)

    # print('Means:', mean)
    # print('Std. Devs:', stddev)

    result = stddev / mean if axis == 0 else numpy.transpose(stddev) / numpy.transpose(mean)

    return result


def coef_correlation(x: numpy.ndarray):
    cov = covariance(x, False)
    stddev = numpy.std(x, axis=0)
    n_obs, n_features = x.shape

    corr = numpy.zeros([n_features, n_features])

    for i in range(n_features):
        for ii in range(i, n_features):
            corr[i, ii] = corr[ii, i] = cov[i, ii] / (stddev[i] * stddev[ii])

    return corr


def covariance(x: numpy.ndarray, minus_one: bool = True):
    """Calculates the covariance matrix of another matrix

    :param x: a data matrix
    :param minus_one: subtract one from the total number of observations

    >>> from sklearn.datasets import load_iris
    >>> x = load_iris()['data']
    >>> print(numpy.array_equal(covariance(x), numpy.cov(x.T)))
    True
    """
    x_means = numpy.mean(x, axis=0)
    n_obs, n_features = x.shape

    covars = numpy.zeros([n_features, n_features])

    for i in range(n_features):
        for ii in range(i, n_features):
            _sum = 0
            for iii in range(n_obs):
                _sum += (x[iii, i] - x_means[i]) * (x[iii, ii] - x_means[ii])
            covars[i, ii] = covars[ii, i] = _sum / ((n_obs - 1) if minus_one else n_obs)

    return covars


def bincount(x):
    """Count the number of occurrences from each integer in a list or 1-D :py:class:`numpy.ndarray`.
    If there are gaps between the numbers, then the numbers in those gaps are given a 0 value of occurrences.

    >>> bincount(numpy.array([-4, 0, 1, 1, 3, 2, 1, 5]))
    array([1, 0, 0, 0, 1, 3, 1, 1, 0, 1], dtype=int32)
    """
    x_max = numpy.max(x)
    x_min = numpy.min(x)
    size = abs(x_max) + abs(x_min) + 1

    count = numpy.zeros(size, dtype=numpy.int32)

    for i in x:
        count[i + abs(x_min)] += 1

    return count


def scatter_matrix(data: numpy.ndarray) -> numpy.ndarray:
    """Calculates the scatter matrix of a data matrix. The scatter matrix is an unnormalized
    version of the covariance matrix, in which the means of the observation values are subtracted.

    The calculations done by this function follow the following equation:

    .. math:: S=\\sum_{{j=1}}^{n}({\\mathbf{x}}_{j}-\\overline
              {{\\mathbf{x}}})({\\mathbf{x}}_{j}-\\overline {{\\mathbf{x}}})^{T}=\\sum
              _{{j=1}}^{n}({\\mathbf{x}}_{j}-\\overline
              {{\\mathbf{x}}})\\otimes({\\mathbf{x}}_{j}-\\overline{{\\mathbf{x}}})=\\left(\\sum
              _{{j=1}}^{n}{\\mathbf {x}}_{j}{\\mathbf {x}}_{j}^{T}\\right)-n\\overline
              {{\\mathbf {x}}}\\overline {{\\mathbf {x}}}^{T}

    :param data: the data matrix
    :returns: the scatter matrix of the given data matrix
    """
    mean_vector = numpy.mean(data, axis=0)
    scatter = numpy.zeros((data.shape[1], data.shape[1]))
    for i in range(data.shape[0]):
        scatter += (data[i, :].reshape(data.shape[1], 1) - mean_vector).dot(
            (data[i, :].reshape(data.shape[1], 1) - mean_vector).T
        )

    return scatter


if __name__ == '__main__':
    import doctest

    doctest.testmod()



import numpy
from scipy.optimize import differential_evolution

from catsim import irt, cat
from catsim.simulation import Estimator


class HillClimbingEstimator(Estimator):
    """Estimator that uses a hill-climbing algorithm to maximize the likelihood function

    :param precision: number of decimal points of precision
    :param verbose: verbosity level of the maximization method
    """

    def __str__(self):
        return 'Hill Climbing Estimator'

    def __init__(self, precision: int = 6, dodd: bool = False, verbose: bool = False):
        super().__init__()
        self._precision = precision
        self._verbose = verbose
        self._evaluations = 0
        self._calls = 0
        self._dodd = dodd

    @property
    def calls(self) -> float:
        """How many times the estimator has been called to maximize/minimize the log-likelihood function

        :returns: number of times the estimator has been called to maximize/minimize the log-likelihood function"""
        return self._calls

    @property
    def evaluations(self) -> float:
        """Total number of times the estimator has evaluated the log-likelihood function during its existence

        :returns: number of function evaluations"""
        return self._evaluations

    @property
    def avg_evaluations(self) -> float:
        """Average number of function evaluations for all tests the estimator has been used

        :returns: average number of function evaluations"""
        return self._evaluations / self._calls

    @property
    def dodd(self) -> bool:
        """Whether Dodd's method will be called by estimator in case the response vector
        is composed solely of right or wrong answers.

        :returns: boolean value indicating if Dodd's method will be used or not."""
        return self._dodd

    def estimate(
        self,
        index: int = None,
        items: numpy.ndarray = None,
        administered_items: list = None,
        response_vector: list = None,
        est_theta: float = None,
        **kwargs
    ) -> float:
        """Returns the theta value that minimizes the negative log-likelihood function, given the current state of the
         test for the given examinee.

        :param index: index of the current examinee in the simulator
        :param items: a matrix containing item parameters in the format that `catsim` understands
                      (see: :py:func:`catsim.cat.generate_item_bank`)
        :param administered_items: a list containing the indexes of items that were already administered
        :param response_vector: a boolean list containing the examinee's answers to the administered items
        :param est_theta: a float containing the current estimated proficiency
        :returns: the current :math:`\\hat\\theta`
        """
        if (index is None or self.simulator is None) and (
            items is None and administered_items is None or response_vector is None or
            est_theta is None
        ):
            raise ValueError(
                'Either pass an index for the simulator or all of the other optional parameters to use this component independently.'
            )

        if items is None and administered_items is None and response_vector is None and est_theta is None:
            items = self.simulator.items
            administered_items = self.simulator.administered_items[index]
            response_vector = self.simulator.response_vectors[index]
            est_theta = self.simulator.latest_estimations[index]

        self._calls += 1
        self._evaluations = 0

        if len(set(response_vector)) == 1 and self._dodd:
            return cat.dodd(est_theta, items, response_vector[-1])

        if set(response_vector) == 1:
            return float('inf')
        elif set(response_vector) == 0:
            return float('-inf')

        if len(administered_items) > 0:
            lower_bound = min(items[administered_items][:, 1])
            upper_bound = max(items[administered_items][:, 1])
        else:
            lower_bound = min(items[:, 1])
            upper_bound = max(items[:, 1])

        best_theta = float('-inf')
        max_ll = float('-inf')
        
        # the estimator starts with a rough search, which gets finer with each pass
        for granularity in range(10):

            # generate a list of candidate theta values
            candidates = numpy.linspace(lower_bound, upper_bound, 10)
            interval_size = candidates[1] - candidates[0]

            if self._verbose:
                print(
                    'Pass: {0}\n\tBounds: {1} {2}\n\tInterval size: {3}'.format(
                        granularity + 1, lower_bound, upper_bound, interval_size
                    )
                )

            # we'll use the concave nature of the log-likelihood function
            # to program a primitive early stopping method in our search
            previous_ll = float('-inf')

            # iterate through each candidate
            for candidate_theta in candidates:
                self._evaluations += 1

                current_ll = irt.log_likelihood(candidate_theta, response_vector, items[administered_items])

                # we search the function from left to right, so when the
                # log-likelihood of the current theta is smaller than the one
                # from the previous theta we tested, it means it's all downhill
                # from then on, so we stop our search
                if current_ll < previous_ll:
                    break
                previous_ll = current_ll

                # check if the LL of the current candidate theta is larger than the best one checked as of yet
                if current_ll > max_ll:
                    if self._verbose:
                        print('\t\tTheta: {0}, LL: {1}'.format(candidate_theta, current_ll))

                    if abs(best_theta - candidate_theta) < float('1e-' + str(self._precision)):
                        return self._getout(candidate_theta)

                    max_ll = current_ll
                    best_theta = candidate_theta

            # the bounds of the new candidates are adjusted around the current best theta value
            lower_bound = best_theta - interval_size
            upper_bound = best_theta + interval_size

        return self._getout(best_theta)

    def _getout(self, theta:float)->float:
        if self._verbose:
            print('{0} evaluations'.format(self._evaluations))

        return theta


class DifferentialEvolutionEstimator(Estimator):
    """Estimator that uses :py:func:`scipy.optimize.differential_evolution` to minimize the negative log-likelihood function

    :param bounds: a tuple containing both lower and upper bounds for the differential
                   evolution algorithm search space. In theory, it is best if they
                   represent the minimum and maximum possible :math:`\\theta` values;
                   in practice, one could also use the smallest and largest difficulty
                   parameters in the item bank, in case no better bounds for
                   :math:`\\theta` exist.
    """

    def __str__(self):
        return 'Differential Evolution Estimator'

    def __init__(self, bounds: tuple):
        super(DifferentialEvolutionEstimator, self).__init__()
        self._lower_bound = min(bounds)
        self._upper_bound = max(bounds)
        self._evaluations = 0
        self._calls = 0

    @property
    def calls(self):
        """How many times the estimator has been called to maximize/minimize the log-likelihood function

        :returns: number of times the estimator has been called to maximize/minimize the log-likelihood function"""
        return self._calls

    @property
    def evaluations(self):
        """Total number of times the estimator has evaluated the log-likelihood function during its existence

        :returns: number of function evaluations"""
        return self._evaluations

    @property
    def avg_evaluations(self):
        """Average number of function evaluations for all tests the estimator has been used

        :returns: average number of function evaluations"""
        return self._evaluations / self._calls

    def estimate(
        self,
        index: int = None,
        items: numpy.ndarray = None,
        administered_items: list = None,
        response_vector: list = None,
        **kwargs
    ) -> float:
        """Uses :py:func:`scipy.optimize.differential_evolution` to return the theta value
        that minimizes the negative log-likelihood function, given the current state of the
        test for the given examinee.

        :param index: index of the current examinee in the simulator
        :param items: a matrix containing item parameters in the format that `catsim` understands
                      (see: :py:func:`catsim.cat.generate_item_bank`)
        :param administered_items: a list containing the indexes of items that were already administered
        :param response_vector: a boolean list containing the examinee's answers to the administered items
        :returns: the current :math:`\\hat\\theta`
        """
        if (index is None or self.simulator is None
            ) and (items is None and administered_items is None or response_vector is None):
            raise ValueError(
                'Either pass an index for the simulator or all of the other optional parameters to use this component independently.'
            )

        if items is None and administered_items is None and response_vector is None:
            items = self.simulator.items
            administered_items = self.simulator.administered_items[index]
            response_vector = self.simulator.response_vectors[index]

        self._calls += 1

        res = differential_evolution(
            irt.negative_log_likelihood,
            bounds=[[self._lower_bound * 2, self._upper_bound * 2]],
            args=(response_vector, items[administered_items])
        )

        self._evaluations = res.nfev

        return res.x[0]



"""Module containing functions relevant to the process of simulating the
application of adaptive tests. Most of this module is based on the work of
[Bar10]_."""

import datetime
import time
import tqdm
from abc import ABCMeta, abstractmethod

import numpy

from . import cat, irt


class Simulable(metaclass=ABCMeta):
    """Base class representing one of the Simulator components that will receive a reference back to it."""

    def __init__(self):
        super(Simulable).__init__()
        self._simulator = None

    @property
    def simulator(self):
        if self._simulator is not None and type(self._simulator) is not Simulator:
            raise ValueError('simulator has to be of type catsim.simulation.Simulator')
        return self._simulator

    @simulator.setter
    def simulator(self, x: 'Simulator'):
        if type(x) is not Simulator:
            raise ValueError('simulator has to be of type catsim.simulation.Simulator')
        self._simulator = x
        self.preprocess()

    def preprocess(self):
        """Override this method to initialize any static values the `Simulable` might use for the duration of the
        simulation. `preprocess` is called after a value is set for the `simulator` property. If a new value if
        attributed to `simulator`, this method is called again, guaranteeing that internal properties of the
        `Simulable` are re-initialized as necessary."""
        pass


class Initializer(Simulable, metaclass=ABCMeta):
    """Base class for CAT initializers"""

    def __init__(self):
        super().__init__()

    @abstractmethod
    def initialize(self, index: int) -> float:
        """Selects an examinee's initial :math:`\\theta` value

        :param index: the index of the current examinee
        :returns: examinee's initial :math:`\\theta` value
        """
        pass


class Selector(Simulable, metaclass=ABCMeta):
    """Base class representing a CAT item selector."""

    def __init__(self):
        super().__init__()

    @abstractmethod
    def select(self, index: int) -> int:
        """Returns the index of the next item to be administered.

        :param index: the index of the current examinee in the simulator.
        :returns: index of the next item to be applied or `None` if there are no more items to be presented.
        """
        pass


class FiniteSelector(Selector, metaclass=ABCMeta):
    """Base class representing a CAT item selector."""

    def __init__(self, test_size):
        self._test_size = test_size
        self._overlap_rate = None
        super().__init__()

    @property
    def test_size(self) -> int:
        return self._test_size

    @property
    def overlap_rate(self) -> float:
        return self._overlap_rate


class Estimator(Simulable, metaclass=ABCMeta):
    """Base class for proficiency estimators"""

    def __init__(self):
        super().__init__()

    @abstractmethod
    def estimate(self, index: int) -> float:
        """Returns the theta value that minimizes the negative log-likelihood function, given the current state of the
         test for the given examinee.

        :param index: index of the current examinee in the simulator
        :returns: the current :math:`\\hat\\theta`
        """
        pass


class Stopper(Simulable, metaclass=ABCMeta):
    """Base class for CAT stop criterion"""

    def __init__(self):
        super().__init__()

    @abstractmethod
    def stop(self, index: int) -> bool:
        """Checks whether the test reached its stopping criterion for the given user

        :param index: the index of the current examinee
        :returns: `True` if the test met its stopping criterion, else `False`"""
        pass


class Simulator:
    """Class representing the simulator. It gathers several objects that describe the full
    simulation process and simulates one or more computerized adaptive tests

    :param items: a matrix containing item parameters
    :param examinees: an integer with the number of examinees, whose real :math:`\\theta` values will be
                      sampled from a normal distribution; or a :py:type:list containing said
                      :math:`\\theta_0` values
    """

    def __init__(
        self,
        items: numpy.ndarray,
        examinees,
        initializer: Initializer = None,
        selector: Selector = None,
        estimator: Estimator = None,
        stopper: Stopper = None
    ):
        irt.validate_item_bank(items)

        # adds a column for each item's exposure rate
        if items.shape[1] < 5:
            items = numpy.append(items, numpy.zeros([items.shape[0], 1]), axis=1)

        self._duration = 0
        self._items = items

        self._bias = None
        self._mse = None
        self._rmse = None
        self._overlap_rate = None

        self._initializer = initializer
        self._selector = selector
        self._estimator = estimator
        self._stopper = stopper

        # `examinees` is passed to its special setter
        self.examinees = examinees

        self._estimations = [[] for _ in range(len(self.examinees))]
        self._administered_items = [[] for _ in range(len(self.examinees))]
        self._response_vectors = [[] for _ in range(len(self.examinees))]

    @property
    def items(self) -> numpy.ndarray:
        """Item matrix used by the simulator. If the simulation already
        occurred, a column containing item exposure rates will be added to the
        matrix."""
        return self._items

    @property
    def administered_items(self) -> list:
        """List of lists containing the indexes of items administered to each
        examinee during the simulation."""
        return self._administered_items

    @property
    def estimations(self) -> list:
        """List of lists containing all estimated :math:`\\hat\\theta` values
        for all examinees during each step of the test."""
        return self._estimations

    @property
    def response_vectors(self) -> list:
        """List of boolean lists containing the examinees answers to all items."""
        return self._response_vectors

    @property
    def latest_estimations(self) -> list:
        """Final estimated :math:`\\hat\\theta` values for all examinees."""
        return [ests[-1] if len(ests) > 0 else None for ests in self._estimations]

    @property
    def examinees(self) -> list:
        """List containing examinees true proficiency values (:math:`\\theta`)."""
        return self._examinees

    @property
    def duration(self) -> float:
        """Duration of the simulation, in seconds."""
        return self._duration

    @property
    def overlap_rate(self) -> float:
        """Overlap rate of the test, if it is of finite length."""
        return self._overlap_rate

    @property
    def initializer(self) -> Initializer:
        return self._initializer

    @property
    def selector(self) -> Selector:
        return self._selector

    @property
    def estimator(self) -> Estimator:
        return self._estimator

    @property
    def stopper(self) -> Stopper:
        return self._stopper

    @property
    def bias(self) -> float:
        """Bias between the estimated and true proficiencies. This property is only
        available after :py:func:`simulate` has been successfully called. For more
        information on estimation bias, see :py:func:`catsim.cat.bias`"""
        return self._bias

    @property
    def mse(self) -> float:
        """Mean-squared error between the estimated and true proficiencies. This
        property is only available after :py:func:`simulate` has been successfully
        called. For more information on the mean-squared error of estimation, see
        :py:func:`catsim.cat.mse`"""
        return self._mse

    @property
    def rmse(self) -> float:
        """Root mean-squared error between the estimated and true proficiencies. This
        property is only available after :py:func:`simulate` has been successfully
        called. For more information on the root mean-squared error of estimation, see
        :py:func:`catsim.cat.rmse`"""
        return self._rmse

    @examinees.setter
    def examinees(self, x):
        # generate examinees from a normal distribution
        # if an int was passed
        if type(x) == int:
            if self._items is not None:
                mean = numpy.mean(self._items[:, 1])
                stddev = numpy.std(self._items[:, 1])
                self._examinees = numpy.random.normal(mean, stddev, x)
            else:
                self._examinees = numpy.random.normal(0, 1, x)
        elif type(x) == list:
            self._examinees = numpy.array(x)
        elif type(x) == numpy.ndarray and x.ndim == 1:
            self._examinees = x
        else:
            raise ValueError(
                'Examinees must be an int, list of floats or one-dimensional numpy array'
            )

    def simulate(
        self,
        initializer: Initializer = None,
        selector: Selector = None,
        estimator: Estimator = None,
        stopper: Stopper = None,
        verbose: bool = False
    ):
        """Simulates a computerized adaptive testing application to one or more examinees

        :param initializer: an initializer that selects examinees :math:`\\theta_0`
        :param selector: a selector that selects new items to be presented to examinees
        :param estimator: an estimator that reestimates examinees proficiencies after each item is applied
        :param stopper: an object with a stopping criteria for the test
        :param verbose: whether to periodically print a message regarding the progress of the simulation.
                        Good for longer simulations.

        >>> from catsim.initialization import RandomInitializer
        >>> from catsim.selection import MaxInfoSelector
        >>> from catsim.estimation import HillClimbingEstimator
        >>> from catsim.stopping import MaxItemStopper
        >>> from catsim.simulation import Simulator
        >>> from catsim.cat import generate_item_bank
        >>> initializer = RandomInitializer()
        >>> selector = MaxInfoSelector()
        >>> estimator = HillClimbingEstimator()
        >>> stopper = MaxItemStopper(20)
        >>> Simulator(generate_item_bank(100), 10).simulate(initializer, selector, estimator, stopper)
        """

        if initializer is not None:
            self._initializer = initializer
        if selector is not None:
            self._selector = selector
        if estimator is not None:
            self._estimator = estimator
        if stopper is not None:
            self._stopper = stopper

        for s in [self._initializer, self._selector, self._estimator, self._stopper]:
            s.simulator = self

        if verbose:
            print(
                (
                    'Starting simulation: {0} {1} {2} {3} {4} items'.format(
                        self._initializer, self._selector, self._estimator, self._stopper,
                        self._items.shape[0]
                    )
                )
            )
            pbar = tqdm.tqdm(total=len(self.examinees))

        start_time = time.time()

        for current_examinee, true_theta in enumerate(self.examinees):

            if verbose:
                pbar.update()

            est_theta = self._initializer.initialize(current_examinee)
            self._estimations[current_examinee].append(est_theta)

            while not self._stopper.stop(current_examinee):
                selected_item = self._selector.select(current_examinee)

                # if the selector returns None, it means the selector and not the stopper, is asking the test to stop
                # this happens e.g. if the item bank or or the available strata end before the minimum error is achieved
                if selected_item is None:
                    break

                # simulates the examinee's response via the four-parameter
                # logistic function
                response = irt.icc(
                    true_theta, self.items[selected_item][0], self.items[selected_item][1],
                    self.items[selected_item][2], self.items[selected_item][3]
                ) >= numpy.random.uniform()

                self._response_vectors[current_examinee].append(response)

                # adds the item selected by the selector to the pool of administered items
                self._administered_items[current_examinee].append(selected_item)

                # estimate the new theta using the given estimator
                est_theta = self._estimator.estimate(current_examinee)

                # count occurrences of this item in all tests
                item_occurrences = numpy.sum(
                    [
                        selected_item in administered_list
                        for administered_list in self._administered_items
                    ]
                )

                # update the exposure value for this item
                # r = number of tests item has been used on / total number of tests
                self.items[selected_item, 4] = item_occurrences / len(self.examinees)

                self._estimations[current_examinee].append(est_theta)

        self._duration = (time.time() - start_time)

        if verbose:
            pbar.close()
            print('Simulation took {0} seconds'.format(self._duration))

        self._bias = cat.bias(self.examinees, self.latest_estimations)
        self._mse = cat.mse(self.examinees, self.latest_estimations)
        self._rmse = cat.rmse(self.examinees, self.latest_estimations)

        # overlap is calculated only if all examinees answered the same amount of items
        # maybe there is a way to calculate it with tests of different lengths,
        # but I did not find it in the literature
        test_size = None
        len_first = len(self._administered_items[0]) if self._administered_items else None
        if type(selector) is FiniteSelector:
            test_size = selector.test_size
        elif all(len(i) == len_first for i in self._administered_items):
            test_size = len_first
        if test_size is not None:
            self._overlap_rate = cat.overlap_rate(self.items[:, 4], test_size)


if __name__ == '__main__':
    import doctest

    doctest.testmod()



import math

import numexpr
import numpy
from numpy import pi


def icc(theta: float, a: float, b: float, c: float = 0, d: float = 1) -> float:
    """Item Response Theory four-parameter logistic function [Ayala2009]_, [Magis13]_:

    .. math:: P(X_i = 1| \\theta) = c_i + \\frac{d_i-c_i}{1+ e^{a_i(\\theta-b_i)}}

    :param theta: the individual's proficiency value. This parameter value has
                  no boundary, but if a distribution of the form :math:`N(0, 1)` was
                  used to estimate the parameters, then :math:`-4 \\leq \\theta \\leq
                  4`.

    :param a: the discrimination parameter of the item, usually a positive
              value in which :math:`0.8 \\leq a \\leq 2.5`.

    :param b: the item difficulty parameter. This parameter value has no
              boundaries, but it is necessary that it be in the same value space
              as `theta` (usually :math:`-4 \\leq b \\leq 4`).

    :param c: the item pseudo-guessing parameter. Being a probability,
              :math:`0\\leq c \\leq 1`, but items considered good usually have
              :math:`c \\leq 0.2`.

    :param d: the item upper asymptote. Being a probability,
              :math:`0\\leq d \\leq 1`, but items considered good usually have
              :math:`d \\approx 1`.
    """
    return c + ((d - c) / (1 + math.e**(-a * (theta - b))))


def _split_params(items: numpy.ndarray):
    """Split the item matrix parameters into columns.

    :param items: an item matrix with four columns representing four parameters.
    :returns: a 4-tuple with each column in a different slot."""
    return items[:, 0], items[:, 1], items[:, 2], items[:, 3]


def detect_model(items: numpy.ndarray) -> int:
    """Detects which logistic model an item matrix fits into.
    :param items: an item matrix
    :return: an int between 1 and 4 denoting the logistic model of the given item matrix
    """
    a, b, c, d = _split_params(items)

    if any(d != 1): return 4
    if any(c != 0): return 3
    if len(set(a)) == 1: return 2
    return 1


def icc_hpc(theta: float, items: numpy.ndarray) -> numpy.ndarray:
    """Implementation of :py:func:`icc` using :py:mod:`numpy` and :py:mod:`numexpr` in which the characteristic
    function for all items in a `numpy.ndarray` are calculated at once.

    :param theta: the individual's proficiency value.
    :param items: array containing the four item parameters.
    :returns: an array of all item characteristic functions, given the current ``theta``"""
    a, b, c, d = _split_params(items)

    return numexpr.evaluate('c + ((d - c) / (1 + exp((-a * (theta - b)))))')


def inf_hpc(theta: float, items: numpy.ndarray):
    """Implementation of :py:func:`inf` using :py:mod:`numpy` and :py:mod:`numexpr` in which the information
    value for all items in a `numpy.ndarray` are calculated at once.

    :param theta: the individual's proficiency value.
    :param items: array containing the four item parameters.
    :returns: an array of all item information values, given the current ``theta``"""
    a, b, c, d = _split_params(items)
    p = icc_hpc(theta, items)

    return numexpr.evaluate('(a ** 2 * (p - c) ** 2 * (d - p) ** 2) / ((d - c) ** 2 * p * (1 - p))')


def inf(theta: float, a: float, b: float, c: float = 0, d: float = 1) -> float:
    """Calculates the information value of an item using the Item Response Theory
    four-parameter logistic model function [Ayala2009]_, [Magis13]_:

    .. math:: I_i(\\theta) = \\frac{a^2[(P(\\theta)-c)]^2[d - P(\\theta)]^2}{(d-c)^2[1-P(\\theta)]P(\\theta)}

    :param theta: the individual's proficiency value. This parameter value has
                  no boundary, but if a distribution of the form
                  :math:`N(0, 1)` was used to estimate the parameters, then
                  :math:`-4 \\leq \\theta \\leq 4`.

    :param a: the discrimination parameter of the item, usually a positive
              value in which :math:`0.8 \\leq a \\leq 2.5`.

    :param b: the item difficulty parameter. This parameter value has no
              boundary, but if a distribution of the form :math:`N(0, 1)` was
              used to estimate the parameters, then :math:`-4 \\leq b \\leq 4`.

    :param c: the item pseudo-guessing parameter. Being a probability,
              :math:`0\\leq c \\leq 1`, but items considered good usually have
              :math:`c \\leq 0.2`.

    :param d: the item upper asymptote. Being a probability,
              :math:`0\\leq d \\leq 1`, but items considered good usually have
              :math:`d \\approx 1`.

    :returns: the information value of the item at the designated `theta` point."""
    p = icc(theta, a, b, c, d)

    return (a**2 * (p - c)**2 * (d - p)**2) / ((d - c)**2 * p * (1 - p))


def test_info(theta: float, items: numpy.ndarray) -> float:
    """Computes the test information of a test at a specific :math:`\\theta` value [Ayala2009]_:

    .. math:: I(\\theta) = \\sum_{j \\in J} I_j(\\theta)

    where :math:`J` is the set of items in the test and :math:`I_j(\\theta)` is the
    item information of :math:`j` at aspecific :math:`\\theta` value.

    :param theta: a proficiency value.
    :param items: a matrix containing item parameters.
    :returns: the test information at `theta` for a test represented by `items`.
    """
    return numpy.sum(inf_hpc(theta, items))


def var(theta: float, items: numpy.ndarray) -> float:
    """Computes the variance (:math:`Var`) of the proficiency estimate of a test at a
    specific :math:`\\theta` value [Ayala2009]_:

    .. math:: Var = \\frac{1}{I(\\theta)}

    where :math:`I(\\theta)` is the test information (see :py:func:`test_info`).

    :param theta: a proficiency value.
    :param items: a matrix containing item parameters.
    :returns: the variance of proficiency estimation at `theta` for a test represented by `items`.
    """
    try:
        return 1 / test_info(theta, items)
    except ZeroDivisionError:
        return float('-inf')


def see(theta: float, items: numpy.ndarray) -> float:
    """Computes the standard error of estimation (:math:`SEE`) of a test at a
    specific :math:`\\theta` value [Ayala2009]_:

    .. math:: SEE = \\sqrt{\\frac{1}{I(\\theta)}}

    where :math:`I(\\theta)` is the test information (see :py:func:`test_info`).

    :param theta: a proficiency value.
    :param items: a matrix containing item parameters.
    :returns: the standard error of estimation at `theta` for a test represented by `items`.
    """
    try:
        return math.sqrt(var(theta, items))
    except ValueError:
        return float('inf')


def reliability(theta: float, items: numpy.ndarray):
    """ Computes test reliability [Thissen00]_, given by:

    .. math:: Rel = 1 - \\frac{1}{I(\\theta)}

    Test reliability is a measure of internal consistency for the test, similar
    to Cronbach's :math:`\\alpha` in Classical Test Theory. Its value is always
    lower than 1, with values close to 1 indicating good reliability. If
    :math:`I(\\theta) < 1`, :math:`Rel < 0` and in these cases it does not make
    sense, but usually the application of additional items solves this problem.

    :param theta: a proficiency value.
    :param items: a matrix containing item parameters.
    :returns: the test reliability at `theta` for a test represented by `items`.
    """
    return 1 - var(theta, items)


def max_info(a: float = 1, b: float = 0, c: float = 0, d: float = 1) -> float:
    """Returns the :math:`\\theta` value to which the item with the given parameters
    gives maximum information. For the 1-parameter and 2-parameter
    logistic models, this :math:`\\theta` corresponds to where :math:`b = 0.5`.
    In the 3-parameter and 4-parameter logistic models, however, this value is given
    by ([Magis13]_)

    .. math:: argmax_{\\theta}I(\\theta) = b + \\frac{1}{a} log \\left(\\frac{x^* - c}{d - x^*}\\right)

    where

    .. math:: x^* = 2 \\sqrt{\\frac{-u}{3}} cos\\left\{\\frac{1}{3}acos\\left(-\\frac{v}{2}\\sqrt{\\frac{27}{-u^3}}\\right)+\\frac{4 \\pi}{3}\\right\} + 0.5

    .. math:: u = -\\frac{3}{4} + \\frac{c + d - 2cd}{2}

    .. math:: v = -\\frac{c + d - 1}{4}

    A few results can be seen in the plots below:

    .. plot::

        from catsim.cat import generate_item_bank
        from catsim import plot
        items = generate_item_bank(2)
        for item in items:
            plot.item_curve(item[0], item[1], item[2], item[3], ptype='iic', max_info=True)

    :param a: item discrimination parameter
    :param b: item difficulty parameter
    :param c: item pseudo-guessing parameter
    :param d: item upper asymptote
    :returns: the :math:`\\theta` value to which the item with the given parameters
              gives maximum information
    """
    # for explanations on finding the following values, see referenced work in function description
    if d == 1:
        if c == 0:
            return b
        return b + (1 / a) * math.log((1 + math.sqrt(1 + 8 * c)) / 2)
    else:
        u = -(3 / 4) + ((c + d - 2 * c * d) / 2)
        v = (c + d - 1) / 4
        x_star = 2 * math.sqrt(-u / 3) * math.cos(
            (1 / 3) * math.acos(-(v / 2) * math.sqrt(27 / (-math.pow(u, 3)))) + (4 * math.pi / 3)
        ) + 0.5

        return b + (1 / a) * math.log((x_star - c) / (d - x_star))


def max_info_hpc(items: numpy.ndarray):
    """Implementation of :py:func:`max_info` using :py:mod:`numpy` and :py:mod:`numexpr` in which the theta
    value that maximizes the information function for all items is returned in a `numpy.ndarray`.

    :param items: array containing the four item parameters.
    :returns: an array of all theta values that maximize the information function of each item."""
    a, b, c, d = _split_params(items)

    if all(d == 1):
        if all(c == 0):
            return b
        return numexpr.evaluate('b + (1 / a) * log((1 + sqrt(1 + 8 * c)) / 2)')
    else:
        u = numexpr.evaluate('-(3 / 4) + ((c + d - 2 * c * d) / 2)')
        v = numexpr.evaluate('(c + d - 1) / 4')
        x_star = numexpr.evaluate(
            '2 * sqrt(-u / 3) * cos((1 / 3) * arccos(-(v / 2) * sqrt(27 / -(u ** 3))) + (4 * pi / 3)) + 0.5'
        )

        return numexpr.evaluate('b + (1 / a) * log((x_star - c) / (d - x_star))')


def log_likelihood(
    est_theta: float, response_vector: list, administered_items: numpy.ndarray
) -> float:
    """Calculates the log-likelihood of an estimated proficiency, given a
    response vector and the parameters of the answered items [Ayala2009]_.

    The likelihood function of a given :math:`\\theta` value given the answers to :math:`I` items is given by:

    .. math:: L(X_{Ij} | \\theta_j, a_I, b_I, c_I, d_I) = \\prod_{i=1} ^ I P_{ij}(\\theta)^{X_{ij}} Q_{ij}(\\theta)^{1-X_{ij}}

    For mathematical reasons, finding the maximum of :math:`L(X_{Ij}` includes using the
    product rule of derivations. Since :math:`L(X_{Ij}` has :math:`j` parts, it can be quite
    complicated to do so. Also, for computational reasons, the product of probabilities can
    quickly tend to 0, so it is common to use the log-likelihood in maximization/minimization
    problems, transforming the product of probabilities in a sum of probabilities:

     .. math:: \\log L(X_{Ij} | \\theta_j, a_I, b_I, c_I, d_I) = \\sum_{i=1} ^ I
               \\left\\lbrace x_{ij} \\log P_{ij}(\\theta)+ (1 - x_{ij}) \\log
               Q_{ij}(\\theta) \\right\\rbrace

    :param est_theta: estimated proficiency value.
    :param response_vector: a Boolean list containing the response vector.
    :param administered_items: a numpy array containing the parameters of the answered items.
    :returns: log-likelihood of a given proficiency value, given the responses to the administered items.
    """
    if len(response_vector) != administered_items.shape[0]:
        raise ValueError(
            'Response vector and administered items must have the same number of items'
        )
    if len(set(response_vector) - {True, False}) > 0:
        raise ValueError('Response vector must contain only Boolean elements')

    ps = icc_hpc(est_theta, administered_items)
    ll = numexpr.evaluate('sum(where(response_vector, log(ps), log(1 - ps)))')

    return ll


def negative_log_likelihood(est_theta: float, *args) -> float:
    """Function used by :py:mod:`scipy.optimize` optimization functions that tend to minimize
    values, instead of maximizing them. Calculates the negative log-likelihood of a proficiency
    value, given a response vector and the parameters of the administered items. The value of
    :py:func:`negative_log_likelihood` is simply the value of :math:`-` :py:func:`log_likelihood` or, mathematically:

    .. math:: - \\log L(X_{Ij} | \\theta_j, a_I, b_I, c_I, d_I)

    :param est_theta: estimated proficiency value

    args:

    :param response_vector list: a Boolean list containing the response vector
    :param administered_items numpy.ndarray: a numpy array containing the parameters of the answered items
    :returns: negative log-likelihood of a given proficiency value, given the responses to the administered items
    """
    return -log_likelihood(est_theta, args[0], args[1])


def normalize_item_bank(items: numpy.ndarray) -> numpy.ndarray:
    """Normalize an item matrix so that it conforms to the standard used by catsim.
    The item matrix must have dimension nx3, in which column 1 represents item discrimination,
    column 2 represents item difficulty, column 3 represents the pseudo-guessing parameter and
    column 4 represents the item upper asymptote.

    If the matrix has one column, it is assumed to be the difficulty column and the other
    two columns are added such that items simulate the 1-parameter logistic model.

    If the matrix has two columns, they are assumed to be the discrimination and difficulty
    columns, respectively. The pseudo-guessing column is added such that items simulate
    the 2-parameter logistic model.

    If the matrix has three columns, they are assumed to be the discrimination, difficulty
    and pseudo-guessing columns, respectively. The upper asymptote column is added such that
    items simulate the 3-parameter logistic model.

    :param items: the item matrix.
    :returns: an nx4 item matrix conforming to the 4 parameter logistic model.
    """
    if len(items.shape) == 1:
        items = numpy.expand_dims(items, axis=0)
    if items.shape[1] == 1:
        items = numpy.append(numpy.ones((items.shape[0])), items, axis=1)
    if items.shape[1] == 2:
        items = numpy.append(items, numpy.zeros((items.shape[0])), axis=1)
    if items.shape[1] == 3:
        items = numpy.append(items, numpy.ones((items.shape[0])), axis=1)

    return items


def validate_item_bank(items: numpy.ndarray, raise_err: bool = False):
    """Validates the shape and parameters in the item matrix so that it conforms to the standard
    used by catsim. The item matrix must have dimension nx4, in which column 1 represents item
    discrimination, column 2 represents item difficulty, column 3 represents the
    pseudo-guessing parameter and column 4 represents the item upper asymptote.

    The item matrix must have at least one line, exactly four columns and
    :math:`\\forall i \\in I , a_i > 0 \\wedge 0 < c_i < 1 \\wedge 0 < d_i < 1`

    :param items: the item matrix.
    :param raise_err: whether to raise an error in case the validation fails or
                      just print the error message to standard output.
    """
    if type(items) is not numpy.ndarray:
        raise ValueError('Item matrix is not of type {0}'.format(numpy.ndarray))

    err = ''

    if len(items.shape) == 1:
        err += 'Item matrix has only one dimension.'
    elif items.shape[1] > 4:
        print(
            '\nItem matrix has more than 4 columns. catsim tends to add \
            columns to the matrix during the simulation, so it\'s not a good idea to keep them.'
        )
    elif items.shape[1] < 4:
        if items.shape[1] == 1:
            err += '\nItem matrix has no discrimination, pseudo-guessing or upper asymptote parameter columns'
        elif items.shape[1] == 2:
            err += '\nItem matrix has no pseudo-guessing or upper asymptote parameter columns'
        elif items.shape[1] == 3:
            err += '\nItem matrix has no upper asymptote parameter column'
    else:
        if any(items[:, 0] < 0):
            err += '\nThere are items with discrimination < 0'
        if any(items[:, 2] < 0):
            err += '\nThere are items with pseudo-guessing < 0'
        if any(items[:, 2] > 1):
            err += '\nThere are items with pseudo-guessing > 1'
        if any(items[:, 3] > 1):
            err += '\nThere are items with upper asymptote > 1'
        if any(items[:, 3] < 0):
            err += '\nThere are items with upper asymptote < 0'

    if len(err) > 0 and raise_err:
        raise ValueError(err)

    print(err)



from abc import abstractmethod
from warnings import warn

import numpy
from scipy.integrate import quad

from . import irt
from .simulation import Selector, FiniteSelector


def _nearest(array: list, value) -> numpy.ndarray:
    """Returns the indexes of values in `array` that are closest to `value`
    :param array: an array of numeric values
    :param value: a numerical value
    :return: an array containing the indexes of numbers in `array`,
             according to how close their are to `value`
    """
    array = numpy.asarray(array)
    return numpy.abs(array - value).argsort()


class MaxInfoSelector(Selector):
    """Selector that returns the first non-administered item with maximum information, given an estimated theta
       
    :param r_max: maximum exposure rate for items
    """

    def __init__(self, r_max: float = 1):
        super().__init__()
        self._r_max = r_max

    def __str__(self):
        return 'Maximum Information Selector'

    @property
    def r_max(self):
        return self._r_max

    def select(
        self,
        index: int = None,
        items: numpy.ndarray = None,
        administered_items: list = None,
        est_theta: float = None,
        **kwargs
    ) -> int:
        """Returns the index of the next item to be administered.

        :param index: the index of the current examinee in the simulator.
        :param items: a matrix containing item parameters in the format that `catsim` understands
                      (see: :py:func:`catsim.cat.generate_item_bank`)
        :param administered_items: a list containing the indexes of items that were already administered
        :param est_theta: a float containing the current estimated proficiency
        :returns: index of the next item to be applied or `None` if there are no more items in the item bank.
        """
        if (index is None or self.simulator is None
            ) and (items is None or administered_items is None or est_theta is None):
            raise ValueError(
                'Either pass an index for the simulator or all of the other optional parameters to use this component independently.'
            )

        if items is None and administered_items is None and est_theta is None:
            items = self.simulator.items
            administered_items = self.simulator.administered_items[index]
            est_theta = self.simulator.latest_estimations[index]

        # first, we'll order items by their information value
        if irt.detect_model(items) <= 2:
            # when the logistic model has the number of parameters <= 2,
            # all items have highest information where theta = b
            ordered_items = _nearest(items[:, 1], est_theta)
        else:
            # else, we'll have to calculate the theta value where information is maximum
            inf_values = irt.max_info_hpc(items)
            ordered_items = _nearest(inf_values, est_theta)

        valid_indexes = [x for x in ordered_items if x not in administered_items]
        if len(valid_indexes) == 0:
            warn('There are no more items to be applied.')
            return None

        # gets the indexes and information values from the items with r < rmax
        valid_indexes_low_r = [index for index in valid_indexes if items[index, 4] < self._r_max]

        # return the item with maximum information from the ones available
        if len(valid_indexes_low_r) > 0:
            selected_item = valid_indexes_low_r[0]
        else:
            selected_item = valid_indexes[0]

        return selected_item


class UrrySelector(Selector):
    """Selector that returns the item whose difficulty parameter is closest to the examinee's proficiency"""

    def __init__(self):
        super().__init__()

    def __str__(self):
        return 'Urry Selector'

    def select(
        self,
        index: int = None,
        items: numpy.ndarray = None,
        administered_items: list = None,
        est_theta: float = None,
        **kwargs
    ) -> int:
        """Returns the index of the next item to be administered.

        :param index: the index of the current examinee in the simulator.
        :param items: a matrix containing item parameters in the format that `catsim` understands
                      (see: :py:func:`catsim.cat.generate_item_bank`)
        :param administered_items: a list containing the indexes of items that were already administered
        :param est_theta: a float containing the current estimated proficiency
        :returns: index of the next item to be applied or `None` if there are no more items in the item bank.
        """
        if (index is None or self.simulator is None
            ) and (items is None or administered_items is None or est_theta is None):
            raise ValueError(
                'Either pass an index for the simulator or all of the other optional parameters to use this component independently.'
            )

        if items is None and administered_items is None and est_theta is None:
            items = self.simulator.items
            administered_items = self.simulator.administered_items[index]
            est_theta = self.simulator.latest_estimations[index]

        ordered_items = _nearest(items[:, 1], est_theta)
        valid_indexes = [x for x in ordered_items if x not in administered_items]

        if len(valid_indexes) == 0:
            warn('There are no more items to be applied.')
            return None

        return valid_indexes[0]


class LinearSelector(FiniteSelector):
    """Selector that returns item indexes in a linear order, simulating a standard
    (non-adaptive) test.

    :param indexes: the indexes of the items that will be returned in order"""

    def __str__(self):
        return 'Linear Selector'

    def __init__(self, indexes: list):
        super().__init__(len(indexes))
        self._indexes = indexes
        self._current = 0

    @property
    def indexes(self):
        return self._indexes

    @property
    def current(self):
        return self._current

    def select(self, index: int = None, administered_items: list = None, **kwargs) -> int:
        """Returns the index of the next item to be administered.

        :param index: the index of the current examinee in the simulator.
        :param administered_items: a list containing the indexes of items that were already administered
        :returns: index of the next item to be applied or `None` if there are no more items in the item bank.
        """
        if (index is None or self.simulator is None) and (administered_items is None):
            raise ValueError(
                'Either pass an index for the simulator or all of the other optional parameters to use this component independently.'
            )

        if administered_items is None:
            administered_items = self.simulator.administered_items[index]

        if set(self._indexes) <= set(administered_items):
            warn(
                'A new index was asked for, but there are no more item indexes to present.\nCurrent item:\t\t\t{0}\nItems to be administered:\t{1} (size: {2})\nAdministered items:\t\t{3} (size: {4})'
                .format(
                    self._current, sorted(self._indexes), len(self._indexes),
                    sorted(administered_items), len(administered_items)
                )
            )
            return None

        selected_item = [x for x in self._indexes if x not in administered_items][0]

        return selected_item


class RandomSelector(Selector):
    """Selector that randomly selects items for application.

    :param replace: whether to select an item that has already been selected before for this examinee."""

    def __str__(self):
        return 'Random Selector'

    def __init__(self, replace: bool = False):
        super().__init__()
        self._replace = replace

    def select(
        self,
        index: int = None,
        items: numpy.ndarray = None,
        administered_items: list = None,
        **kwargs
    ) -> int:
        """Returns the index of the next item to be administered.

        :param index: the index of the current examinee in the simulator.
        :param items: a matrix containing item parameters in the format that `catsim` understands
                      (see: :py:func:`catsim.cat.generate_item_bank`)
        :param administered_items: a list containing the indexes of items that were already administered
        :returns: index of the next item to be applied or `None` if there are no more items in the item bank.
        """
        if (index is None or
            self.simulator is None) and (items is None or administered_items is None):
            raise ValueError(
                'Either pass an index for the simulator or all of the other optional parameters to use this component independently.'
            )

        if items is None and administered_items is None:
            items = self.simulator.items
            administered_items = self.simulator.administered_items[index]

        if len(administered_items) >= items.shape[0] and not self._replace:
            warn(
                'A new item was asked for, but there are no more items to present.\nAdministered items:\t{0}\nItem bank size:\t{1}'
                .format(len(administered_items), items.shape[0])
            )
            return None

        if self._replace:
            return numpy.random.choice(items.shape[0])
        else:
            return numpy.random.choice(list(set(range(items.shape[0])) - set(administered_items)))


class ClusterSelector(Selector):
    """Cluster-based Item Selection Method.

        .. [Men15] Meneghetti, D. R. (2015). Metolodogia de seleção de itens em testes
           adaptativos informatizados baseada em agrupamento por similaridade (Mestrado).
           Centro Universitário da FEI. Retrieved from
           https://www.researchgate.net/publication/283944553_Metodologia_de_selecao_de_itens_em_Testes_Adaptativos_Informatizados_baseada_em_Agrupamento_por_Similaridade

    :param clusters: a list containing item cluster memberships
    :param r_max: maximum exposure rate for items
    :param method: one of the available methods for cluster selection. Given
                   the estimated theta value at each step:

                       ``item_info``: selects the cluster which has the item
                       with maximum information;

                       ``cluster_info``: selects the cluster whose items sum of
                       information is maximum;

                       ``weighted_info``: selects the cluster whose weighted
                       sum of information is maximum. The weighted equals the
                       number of items in the cluster;

    :param r_control: if `passive` and all items :math:`i` in the selected
                      cluster have :math:`r_i > r^{max}`, applies the item with
                      maximum information; if `aggressive`, applies the item
                      with smallest :math:`r` value.
    """

    def __str__(self):
        return 'Cluster Selector'

    @property
    def r_max(self):
        return self._r_max

    @property
    def clusters(self):
        return self._clusters

    @property
    def method(self):
        return self._method

    @property
    def r_control(self):
        return self._r_control

    def __init__(
        self,
        clusters: list,
        method: str = 'item_info',
        r_max: float = 1,
        r_control: str = 'passive'
    ):
        super().__init__()
        available_methods = ['item_info', 'cluster_info', 'weighted_info']
        if method not in available_methods:
            raise ValueError(
                '{0} is not a valid cluster selection method; choose one from {1}'.format(
                    method, available_methods
                )
            )
        available_rcontrol = ['passive', 'aggressive']
        if r_control not in available_rcontrol:
            raise ValueError(
                '{0} is not a valid item exposure control method; choose one from {1}'.format(
                    r_control, available_rcontrol
                )
            )

        self._clusters = clusters
        self._method = method
        self._r_max = r_max
        self._r_control = r_control

    def select(
        self,
        index: int = None,
        items: numpy.ndarray = None,
        administered_items: list = None,
        est_theta: float = None,
        **kwargs
    ) -> int:
        """Returns the index of the next item to be administered.

        :param index: the index of the current examinee in the simulator.
        :param items: a matrix containing item parameters in the format that `catsim` understands
                      (see: :py:func:`catsim.cat.generate_item_bank`)
        :param administered_items: a list containing the indexes of items that were already administered
        :param est_theta: a float containing the current estimated proficiency
        :returns: index of the next item to be applied.
        """
        if (index is None or self.simulator is None
            ) and (items is None or administered_items is None or est_theta is None):
            raise ValueError(
                'Either pass an index for the simulator or all of the other optional parameters to use this component independently.'
            )

        if items is None and administered_items is None and est_theta is None:
            items = self.simulator.items
            administered_items = self.simulator.administered_items[index]
            est_theta = self.simulator.latest_estimations[index]

        selected_cluster = None
        existent_clusters = set(self._clusters)

        # this part of the code selects the cluster from which the item at
        # the current point of the test will be chosen
        if self._method == 'item_info':
            # get the item indexes sorted by their information value
            infos = _nearest(irt.max_info_hpc(items), est_theta)

            evaluated_clusters = set()

            # iterate over every item in order of information value
            for i in range(items.shape[0]):
                # get the current non-examined item
                max_info_item = infos[i]

                # if the cluster of the current item has already been fully examined, go to the next item
                if self._clusters[max_info_item] in evaluated_clusters:
                    continue

                # get the indexes of all items in the same cluster as the current item
                items_in_cluster = numpy.nonzero(
                    [x == self._clusters[max_info_item] for x in self._clusters]
                )[0]

                # if all the items in the current cluster have already been administered (it happens, theoretically),
                # add this cluster to the list of fully evaluated clusters
                if set(items_in_cluster) <= set(administered_items):
                    evaluated_clusters.add(self._clusters[max_info_item])

                    # if all clusters have been evaluated, the loop ends and the method returns None somewhere below
                    if evaluated_clusters == existent_clusters:
                        break

                    # else, if there are still items and clusters to be explored, keep going
                    continue

                # if the algorithm gets here, it means this cluster can be used
                selected_cluster = self._clusters[max_info_item]
                break

        elif self._method in ['cluster_info', 'weighted_info']:
            # calculates the cluster information, depending on the method
            # selected
            if self._method == 'cluster_info':
                cluster_infos = ClusterSelector.sum_cluster_infos(est_theta, items, self._clusters)
            else:
                cluster_infos = ClusterSelector.weighted_cluster_infos(
                    est_theta, items, self._clusters
                )

            # sorts clusters descending by their information values
            # this type of sorting was seem on
            # http://stackoverflow.com/a/6618543
            sorted_clusters = numpy.array(
                [
                    cluster for (inf_value, cluster) in sorted(
                        zip(cluster_infos, set(self._clusters)),
                        key=lambda pair: pair[0],
                        reverse=True
                    )
                ],
                dtype=float
            )

            # walks through the sorted clusters in order
            for i in range(len(sorted_clusters)):
                valid_indexes = numpy.nonzero([r == sorted_clusters[i] for r in items[:, 4]])[0]

                # checks if at least one item from this cluster has not
                # been administered to this examinee yet
                if set(valid_indexes).intersection(administered_items) != set(valid_indexes):
                    selected_cluster = sorted_clusters[i]
                    break
                    # the for loop ends with the cluster that has a) the maximum
                    # information possible and b) at least one item that has not
                    # yet been administered

        # if the test size gets larger than the item bank size, end the test
        if selected_cluster is None:
            warn("There are no more items to be applied.")
            return None

        # in this part, an item is chosen from the cluster that was
        # selected above

        # gets the indexes and information values from the items in the
        # selected cluster that have not been administered
        valid_indexes = [
            index
            for index in numpy.nonzero([cluster == selected_cluster
                                        for cluster in self._clusters])[0]
            if index not in administered_items
        ]

        # gets the indexes and information values from the items in the
        # selected cluster with r < rmax that have not been
        # administered
        valid_indexes_low_r = [
            index for index in valid_indexes
            if items[index, 4] < self._r_max and index not in administered_items
        ]

        if len(valid_indexes_low_r) > 0:
            # return the item with maximum information from the ones available
            inf_values = irt.inf_hpc(est_theta, items[valid_indexes_low_r])
            selected_item = valid_indexes_low_r[numpy.nonzero(inf_values == max(inf_values))[0][0]]

        # if all items in the selected cluster have exceed their r values,
        # select the one with smallest r, regardless of information
        else:
            if self._r_control == 'passive':
                inf_values = irt.inf_hpc(est_theta, items[valid_indexes])
                selected_item = valid_indexes[numpy.nonzero(inf_values == max(inf_values))[0][0]]
            else:
                selected_item = valid_indexes[items[:, 4].index(min(items[:, 4]))]

        return selected_item

    @staticmethod
    def sum_cluster_infos(theta: float, items: numpy.ndarray, clusters: list) -> numpy.ndarray:
        """Returns the sum of item information values, separated by cluster

        :param theta: an examinee's :math:`\\theta` value
        :param items: a matrix containing item parameters in the format that `catsim` understands
                      (see: :py:func:`catsim.cat.generate_item_bank`)
        :param clusters: a list containing item cluster memberships, represented by integers
        :returns: array containing the sum of item information values for each cluster"""
        cluster_infos = numpy.zeros((len(set(clusters))))

        for cluster in set(clusters):
            cluster_indexes = numpy.nonzero([c == cluster for c in clusters])[0]

            for item in items[cluster_indexes]:
                cluster_infos[cluster] += irt.inf(theta, item[0], item[1], item[2], item[3])

        return cluster_infos

    @staticmethod
    def weighted_cluster_infos(theta: float, items: numpy.ndarray, clusters: list) -> numpy.ndarray:
        """Returns the weighted sum of item information values, separated by cluster.
        The weight is the number of items in each cluster.

        :param theta: an examinee's :math:`\\theta` value
        :param items: a matrix containing item parameters in the format that `catsim` understands
                      (see: :py:func:`catsim.cat.generate_item_bank`)
        :param clusters: a list containing item cluster memberships, represented by integers
        :returns: array containing the sum of item information values for each cluster,
                  divided by the number of items in each cluster"""
        cluster_infos = ClusterSelector.sum_cluster_infos(theta, items, clusters)
        count = numpy.bincount(clusters)

        for i in range(len(cluster_infos)):
            cluster_infos[i] /= count[i]

        return cluster_infos

    @staticmethod
    def sum_cluster_params(items: numpy.ndarray, c: list):
        """Returns the sum of item parameter values for each cluster

        :param items: a matrix containing item parameters in the format that `catsim` understands
                      (see: :py:func:`catsim.cat.generate_item_bank`)
        :param c: a list containing clustering memeberships.
        :returns: a matrix containing the sum of each parameter by cluster. Lines are clusters, columns are parameters.
        """
        averages = numpy.zeros((numpy.max(c) + 1, 4))

        for i in numpy.arange(0, numpy.size(c)):
            if c[i] == -1:
                continue
            averages[c[i], 0] += items[i, 0]
            averages[c[i], 1] += items[i, 1]
            averages[c[i], 2] += items[i, 2]
            averages[c[i], 3] += items[i, 3]

        return averages

    @staticmethod
    def avg_cluster_params(items: numpy.ndarray, c: list):
        """Returns the average values of item parameters by cluster

        :param items:
        :param c: a list containing clustering memeberships.
        :returns: a matrix containing the average values of each parameter by cluster.
                  Lines are clusters, columns are parameters."""
        averages = ClusterSelector.sum_cluster_params(items, c)

        occurrences = numpy.bincount(numpy.delete(c, numpy.where(c == -1)).astype(numpy.int64))

        for counter, i in enumerate(occurrences):
            averages[counter, 0] /= i
            averages[counter, 1] /= i
            averages[counter, 2] /= i
            averages[counter, 3] /= i

        return averages


class StratifiedSelector(FiniteSelector):

    def __str__(self):
        return 'General Stratified Selector'

    def __init__(self, test_size):
        super().__init__(test_size)
        self._organized_items = None

    @staticmethod
    @abstractmethod
    def sort_items(items: numpy.ndarray) -> numpy.ndarray:
        pass

    def preprocess(self):
        # sort item indexes by their discrimination value
        self._organized_items = __class__.sort_items(self.simulator.items)

    def select(
        self,
        index: int = None,
        items: numpy.ndarray = None,
        administered_items: list = None,
        **kwargs
    ) -> int:
        """Returns the index of the next item to be administered.

        :param index: the index of the current examinee in the simulator.
        :param items: a matrix containing item parameters
        :param administered_items: a list containing the indexes of items that were already administered
        :returns: index of the next item to be applied or `None` if there are no more strata to get items from.
        """
        if (index is None or
            self.simulator is None) and (items is None or administered_items is None):
            raise ValueError(
                'Either pass an index for the simulator or all of the other optional parameters to use this component independently.'
            )

        if items is None and administered_items is None:
            items = self.simulator.items
            administered_items = self.simulator.administered_items[index]

        # select the item in the correct layer, according to the point in the test the examinee is
        slices = numpy.linspace(0, items.shape[0], self._test_size, endpoint=False, dtype='i')

        try:
            pointer = slices[len(administered_items)]
            max_pointer = items.shape[0] if len(
                administered_items
            ) == self._test_size - 1 else slices[len(administered_items) + 1]
        except IndexError:
            warn(
                "{0}: test size is larger than was informed to the selector\nLength of administered items:\t{0}\nTotal length of the test:\t{1}\nNumber of slices:\t{2}"
                .format(self, len(administered_items), self._test_size, len(slices))
            )
            return None

        organized_items = self._organized_items if self._organized_items is not None else self.sort_items(
            items
        )

        # if the selected item has already been administered, select the next one
        while organized_items[pointer] in administered_items:
            pointer += 1
            if pointer == max_pointer:
                raise ValueError(
                    'There are no more items to be selected from stratum {0}'.format(
                        slices[len(administered_items)]
                    )
                )

        return organized_items[pointer]


class AStratSelector(StratifiedSelector):
    """Implementation of the :math:`\\alpha`-stratified selector proposed by
    [Chang99]_, in which the item bank is sorted in ascending order according to the
    items discrimination parameter and then separated into :math:`K` strata
    (:math:`K` being the test size), each stratum containing gradually higher
    average discrimination. The :math:`\\alpha`-stratified selector then selects the
    first non-administered item from stratum :math:`k`, in which :math:`k`
    represents the position in the test of the current item the examinee is being
    presented.

    .. image:: ../sphinx/alpha-strat.*

    :param test_size: the number of items the test contains. The selector uses this parameter
                      to create the correct number of strata.
    """

    def __str__(self):
        return 'a-Stratified Selector'

    def __init__(self, test_size):
        super().__init__(test_size)

    @staticmethod
    def sort_items(items: numpy.ndarray) -> numpy.ndarray:
        return items[:, 0].argsort()


class AStratBBlockSelector(StratifiedSelector):
    """Implementation of the :math:`\\alpha`-stratified selector with :math:`b`
    blocking proposed by [Chang2001]_, in which the item bank is sorted in ascending
    order according to the items difficulty parameter and then separated into
    :math:`M` strata, each stratum containing gradually higher average difficulty.

    Each of the :math:`M` strata is then again separated into :math:`K`
    sub-strata (:math:`k` being the test size), according to their
    discrimination. The final item bank is then ordered such that the first
    sub-strata of each strata forms the first strata of the new ordered item
    bank, and so on. This method tries to balance the distribution of both
    parameters between all strata, after perceiving that they are correlated.

    .. image:: ../sphinx/b-blocking.*

    :param test_size: the number of items the test contains. The selector uses this parameter to
                      create the correct number of strata.
    """

    def __str__(self):
        return 'a-Stratified b-Blocking Selector'

    def __init__(self, test_size):
        super().__init__(test_size)

    @staticmethod
    def sort_items(items: numpy.ndarray) -> numpy.ndarray:
        return numpy.lexsort((items[:, 0], items[:, 1]))


class MaxInfoStratSelector(StratifiedSelector):
    """Implementation of the maximum information stratification (MIS) selector
    proposed by [Bar06]_, in which the item bank is sorted in ascending order
    according to the items maximum information and then separated into :math:`K`
    strata (:math:`K` being the test size), each stratum containing items with
    gradually higher maximum information. The MIS selector then selects the first
    non-administered item from stratum :math:`k`, in which :math:`k` represents the
    position in the test of the current item the examinee is being presented.

    .. image:: ../sphinx/mis.*

    This method claims to work better than the :math:`a`-stratified method by
    [Chang99]_ for the three-parameter logistic model of IRT, since item difficulty
    and maximum information are not positioned in the same place in the proficiency
    scale in 3PL.

    :param test_size: the number of items the test contains. The selector uses this parameter to
                      create the correct number of strata.
    """

    def __str__(self):
        return 'Maximum Information Stratification Selector'

    def __init__(self, test_size):
        super().__init__(test_size)

    @staticmethod
    def sort_items(items: numpy.ndarray) -> numpy.ndarray:
        maxinfo = irt.max_info_hpc(items)
        return irt.inf_hpc(maxinfo, items).argsort()


class MaxInfoBBlockSelector(StratifiedSelector):
    """Implementation of the maximum information stratification with :math:`b`
    blocking (MIS-B) selector proposed by [Bar06]_, in which the item bank is sorted
    in ascending order according to the items difficulty parameter and then
    separated into :math:`M` strata, each stratum containing gradually higher
    average difficulty.

    Each of the :math:`M` strata is then again separated into :math:`K`
    sub-strata (:math:`k` being the test size), according to the items maximum
    information. The final item bank is then ordered such that the first
    sub-strata of each strata forms the first strata of the new ordered item
    bank, and so on. This method tries to balance the distribution of both
    parameters between all strata and works better than the :math:`a`-stratified
    with :math:`b` blocking method by [Chang2001]_ for the three-parameter
    logistic model of IRT, since item difficulty and maximum information are not
    positioned in the same place in the proficiency scale in 3PL. This may also
    apply, although not mentioned by the authors, for the 4PL.

    .. image:: ../sphinx/mis-b.*

    :param test_size: the number of items the test contains. The selector uses this parameter to
                      create the correct number of strata.
    """

    def __str__(self):
        return 'Maximum Information Stratification with b-Blocking Selector'

    def __init__(self, test_size):
        super().__init__(test_size)

    @staticmethod
    def sort_items(items: numpy.ndarray) -> numpy.ndarray:
        maxinfo = irt.max_info_hpc(items)
        return numpy.lexsort((irt.inf_hpc(maxinfo, items), maxinfo))


class The54321Selector(FiniteSelector):
    """Implementation of the 5-4-3-2-1 selector proposed by [McBride83]_, in which,
    at each step :math:`k` of a test of size :math:`K`, an item is chosen from a bin
    containing the :math:`K-k` most informative items in the bank, given the current
    :math:`\\hat\\theta`. As the test progresses, the bin gets smaller and more
    informative items have a higher probability of being chosen by the end of the
    test, when the estimation of ':math:`\\hat\\theta` is more precise. The
    5-4-3-2-1 selector can be viewed as a specialization of the
    :py:class:`catsim.selection.RandomesqueSelector`, in which the bin size of most
    informative items gets smaller as the test progresses.

    :param test_size: the number of items the test contains. The selector uses
                      this parameter to set the bin size"""

    def __str__(self):
        return '5-4-3-2-1 Selector'

    def __init__(self, test_size):
        super().__init__(test_size)

    def select(
        self,
        index: int = None,
        items: numpy.ndarray = None,
        administered_items: list = None,
        est_theta: float = None,
        **kwargs
    ) -> int:
        """Returns the index of the next item to be administered.

        :param index: the index of the current examinee in the simulator.
        :param items: a matrix containing item parameters in the format that `catsim` understands
                      (see: :py:func:`catsim.cat.generate_item_bank`)
        :param administered_items: a list containing the indexes of items that were already administered
        :param est_theta: a float containing the current estimated proficiency
        :returns: index of the next item to be applied or `None` if there are no more items in the item bank.
        """
        if (index is None or self.simulator is None
            ) and (items is None or administered_items is None or est_theta is None):
            raise ValueError(
                'Either pass an index for the simulator or all of the other optional parameters to use this component independently.'
            )

        if items is None and administered_items is None and est_theta is None:
            items = self.simulator.items
            administered_items = self.simulator.administered_items[index]
            est_theta = self.simulator.latest_estimations[index]

        # sort item indexes by their information value descending and remove indexes of administered items
        organized_items = [
            x for x in (-irt.inf_hpc(est_theta, items)).argsort() if x not in administered_items
        ]

        bin_size = self._test_size - len(administered_items)

        if len(organized_items) == 0:
            warn('There are no more items to apply.')
            return None

        return numpy.random.choice(organized_items[0:bin_size])


class RandomesqueSelector(Selector):
    """Implementation of the randomesque selector proposed by [Kingsbury89]_, in which,
    at every step of the test, an item is randomly chosen from the :math:`n` most informative
    items in the item bank, :math:`n` being a predefined value (originally 5, but user-defined
    in this implementation)

    :param bin_size: the number of most informative items to be taken into consideration when
                     randomly selecting one of them.
    """

    def __str__(self):
        return 'Randomesque Selector'

    def __init__(self, bin_size):
        super().__init__()
        self._bin_size = bin_size

    @property
    def bin_size(self):
        return self._bin_size

    def select(
        self,
        index: int = None,
        items: numpy.ndarray = None,
        administered_items: list = None,
        est_theta: float = None,
        **kwargs
    ) -> int:
        """Returns the index of the next item to be administered.

        :param index: the index of the current examinee in the simulator.
        :param items: a matrix containing item parameters in the format that `catsim` understands
                      (see: :py:func:`catsim.cat.generate_item_bank`)
        :param administered_items: a list containing the indexes of items that were already administered
        :param est_theta: a float containing the current estimated proficiency
        :returns: index of the next item to be applied or `None` if there are no more items in the item bank.
        """
        if (index is None or self.simulator is None
            ) and (items is None or administered_items is None or est_theta is None):
            raise ValueError(
                'Either pass an index for the simulator or all of the other optional parameters to use this component independently.'
            )

        if items is None and administered_items is None and est_theta is None:
            items = self.simulator.items
            administered_items = self.simulator.administered_items[index]
            est_theta = self.simulator.latest_estimations[index]

        # sort item indexes by their information value descending and remove indexes of administered items
        organized_items = [
            x for x in (-irt.inf_hpc(est_theta, items)).argsort() if x not in administered_items
        ]

        if len(organized_items) == 0:
            warn('There are no more items to apply.')
            return None

        return numpy.random.choice(list(organized_items)[:self._bin_size])


class IntervalInfoSelector(Selector):
    """A selector in which, at every step of the test, the item that maximizes
    the integral of the information function at a predetermined ``interval``
    :math:`\\delta` above and below the current :math:`\\hat\\theta` is chosen.

    .. math:: argmax_{i \\in I} \\int_{\\hat\\theta - \\delta}^{\\hat\\theta - \\delta}I_i(\\hat\\theta)

    :param interval: the interval of the integral. If no interval is passed, the
                     integral is calculated from :math:`[-\\infty, \\infty]`.
    """

    def __str__(self):
        return 'Interval Information Selector'

    def __init__(self, interval: float = None):
        super().__init__()
        self._interval = interval if interval is not None else numpy.inf

    @property
    def interval(self):
        return self._interval

    def select(
        self,
        index: int = None,
        items: numpy.ndarray = None,
        administered_items: list = None,
        est_theta: float = None,
        **kwargs
    ) -> int:
        """Returns the index of the next item to be administered.

        :param index: the index of the current examinee in the simulator.
        :param items: a matrix containing item parameters in the format that `catsim` understands
                      (see: :py:func:`catsim.cat.generate_item_bank`)
        :param administered_items: a list containing the indexes of items that were already administered
        :param est_theta: a float containing the current estimated proficiency
        :returns: index of the next item to be applied or `None` if there are no more items in the item bank.
        """
        if (index is None or self.simulator is None
            ) and (items is None or administered_items is None or est_theta is None):
            raise ValueError(
                'Either pass an index for the simulator or all of the other optional parameters to use this component independently.'
            )

        if items is None and administered_items is None and est_theta is None:
            items = self.simulator.items
            administered_items = self.simulator.administered_items[index]
            est_theta = self.simulator.latest_estimations[index]

        # sort item indexes by the integral of the information function descending and remove indexes of administered items
        organized_items = [
            x for x in (-numpy.array([
                    quad(
                        irt.inf,
                        est_theta - self._interval,
                        est_theta + self._interval,
                        args=(item[0], item[1], item[2], item[3])
                    )[0] for item in items]
            )).argsort() if x not in administered_items
        ]

        if len(organized_items) == 0:
            warn('There are no more items to apply.')
            return None

        return list(organized_items)[0]



#!/usr/bin/env python3

from setuptools import setup, find_packages
from catsim import __version__
from os import path

this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'README.rst'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    classifiers=[
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 4 - Beta',
        'Intended Audience :: Education',
        'Intended Audience :: Science/Research',
        'Natural Language :: English',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Topic :: Education :: Testing',
        'License :: OSI Approved :: GNU Lesser General Public License v3 (LGPLv3)'
    ],
    name='catsim',
    version=__version__,
    description='Computerized Adaptive Testing Simulator',
    long_description=long_description,
    long_description_content_type='text/x-rst',
    author='Douglas De Rizzo Meneghetti',
    author_email='douglasrizzom@gmail.com',
    url='https://douglasrizzo.com.br/catsim',
    packages=find_packages(exclude=['contrib', 'docs', 'tests*']),
    install_requires=[
        'scipy', 'numexpr', 'matplotlib', 'scikit-learn', 'json_tricks', 'tqdm', 'numpy'
    ],
    extras_require=dict(
        testing=['nose', 'nose-cov', 'sklearn', 'coverage', 'coveralls', 'python-coveralls'],
        docs=[
            'Sphinx', 'numpydoc', 'sphinx_autodoc_annotation', 'sphinx_rtd_theme',
            'bibtex-pygments-lexer', 'matplotlib'
        ]
    ),
    license='LGPLv3'
)



import numpy

from . import irt
from .simulation import Stopper


class MaxItemStopper(Stopper):
    """Stopping criterion for maximum number of items in a test

    :param max_itens: the maximum number of items in the test"""

    def __str__(self):
        return 'Maximum Item Number Initializer'

    def __init__(self, max_itens: int):
        super(MaxItemStopper, self).__init__()
        self._max_itens = max_itens

    def stop(self, index: int = None, administered_items: numpy.ndarray = None, **kwargs) -> bool:
        """Checks whether the test reached its stopping criterion for the given user

        :param index: the index of the current examinee
        :param administered_items: a matrix containing the parameters of items that were already administered
        :returns: `True` if the test met its stopping criterion, else `False`"""

        if (index is None or self.simulator is None) and administered_items is None:
            raise ValueError

        if administered_items is None:
            administered_items = self.simulator.items[self.simulator.administered_items[index]]

        n_itens = administered_items.shape[0]
        if n_itens > self._max_itens:
            raise ValueError(
                'More items than permitted were administered: {0} > {1}'.format(
                    n_itens, self._max_itens
                )
            )

        return n_itens == self._max_itens


class MinErrorStopper(Stopper):
    """Stopping criterion for minimum standard error of estimation (see :py:func:`catsim.irt.see`)

    :param min_error: the minimum standard error of estimation the test must achieve before stopping"""

    def __str__(self):
        return 'Minimum Error Initializer'

    def __init__(self, min_error: float):
        super(MinErrorStopper, self).__init__()
        self._min_error = min_error

    def stop(
        self,
        index: int = None,
        administered_items: numpy.ndarray = None,
        theta: float = None,
        **kwargs
    ) -> bool:
        """Checks whether the test reached its stopping criterion

        :param index: the index of the current examinee
        :param administered_items: a matrix containing the parameters of items that were already administered
        :param theta: a float containing the a proficiency value to which the error will be calculated
        :returns: `True` if the test met its stopping criterion, else `False`"""

        if (index is None or
            self.simulator is None) and (administered_items is None or theta is None):
            raise ValueError

        if administered_items is None and theta is None:
            theta = self.simulator.latest_estimations[index]
            administered_items = self.simulator.items[self.simulator.administered_items[index]]

        if theta is None:
            return False

        return irt.see(theta, administered_items) < self._min_error



import random
import unittest

from sklearn.cluster import KMeans

from . import stats, plot
from .cat import generate_item_bank
from .estimation import *
from .initialization import *
from .selection import *
from .simulation import Simulator
from .stopping import MaxItemStopper, MinErrorStopper


def one_simulation(items, examinees, initializer, selector, estimator, stopper):
    s = Simulator(items, examinees)
    s.simulate(initializer, selector, estimator, stopper, verbose=True)
    cat.rmse(s.examinees, s.latest_estimations)


def test_cism():
    examinees = 100
    test_sizes = [30]
    bank_sizes = [500]

    for bank_size in bank_sizes:
        for test_size in test_sizes:

            initializers = [RandomInitializer('uniform', (-5, 5))]
            estimators = [HillClimbingEstimator()]
            stoppers = [MaxItemStopper(test_size), MinErrorStopper(.4)]

            for initializer in initializers:
                for estimator in estimators:
                    for stopper in stoppers:
                        items = generate_item_bank(bank_size)
                        clusters = list(KMeans(n_clusters=8).fit_predict(items))
                        ClusterSelector.weighted_cluster_infos(0, items, clusters)
                        ClusterSelector.avg_cluster_params(items, clusters)
                        selector = ClusterSelector(clusters=clusters, r_max=.2)
                        yield one_simulation, items, examinees, initializer, selector, estimator, stopper


def test_simulations():
    examinees = 100
    test_sizes = [30]
    bank_sizes = [500]

    logistic_models = ['4PL']

    for bank_size in bank_sizes:
        for test_size in test_sizes:
            initializers = [RandomInitializer('uniform', (-5, 5)), FixedPointInitializer(0)]
            infinite_selectors = [MaxInfoSelector(), RandomSelector(), UrrySelector()]
            finite_selectors = [
                LinearSelector(list(numpy.random.choice(bank_size, size=test_size, replace=False))),
                AStratSelector(test_size),
                AStratBBlockSelector(test_size),
                MaxInfoStratSelector(test_size),
                MaxInfoBBlockSelector(test_size),
                The54321Selector(test_size),
                RandomesqueSelector(5)
            ]

            for logistic_model in logistic_models:
                for initializer in [FixedPointInitializer(0)]:
                    for estimator in [HillClimbingEstimator()]:
                        for stopper in [MaxItemStopper(test_size)]:
                            for selector in finite_selectors:
                                items = generate_item_bank(bank_size, itemtype=logistic_model)
                                responses = cat.random_response_vector(
                                    random.randint(1, test_size - 1)
                                )
                                administered_items = numpy.random.choice(
                                    bank_size, len(responses), replace=False
                                )
                                est_theta = initializer.initialize()
                                selector.select(
                                    items=items,
                                    administered_items=administered_items,
                                    est_theta=est_theta
                                )
                                estimator.estimate(
                                    items=items,
                                    administered_items=administered_items,
                                    response_vector=responses,
                                    est_theta=est_theta
                                )
                                stopper.stop(
                                    administered_items=items[administered_items], theta=est_theta
                                )

                                yield one_simulation, items, examinees, initializer, selector, estimator, stopper

                        for stopper in [MinErrorStopper(.4), MaxItemStopper(test_size)]:
                            for selector in infinite_selectors:
                                items = generate_item_bank(bank_size, itemtype=logistic_model)
                                yield one_simulation, items, examinees, initializer, selector, estimator, stopper


def test_item_bank_generation():
    for items in [
        generate_item_bank(5, '1PL'),
        generate_item_bank(5, '2PL'),
        generate_item_bank(5, '3PL'),
        generate_item_bank(5, '3PL', corr=0),
        generate_item_bank(5, '4PL')
    ]:
        irt.validate_item_bank(items, raise_err=True)

    items = numpy.zeros(100)
    irt.validate_item_bank(items)
    items = irt.normalize_item_bank(items)
    irt.validate_item_bank(items, raise_err=True)


def test_plots():
    from matplotlib.pyplot import close
    initializer = RandomInitializer()
    selector = MaxInfoSelector()
    estimator = HillClimbingEstimator()
    stopper = MaxItemStopper(20)
    s = Simulator(generate_item_bank(100), 10)
    s.simulate(initializer, selector, estimator, stopper, verbose=True)

    for item in s.items[0:10]:
        yield plot.item_curve, item[0], item[1], item[2], item[
            3
        ], 'Test plot', 'icc', False, None, False
        yield plot.item_curve, item[0], item[1], item[2], item[
            3
        ], 'Test plot', 'iic', True, None, False
        yield plot.item_curve, item[0], item[1], item[2], item[
            3
        ], 'Test plot', 'both', True, None, False
        close('all')

    plot.gen3d_dataset_scatter(items=s.items, show=False)
    plot.test_progress(
        title='Test progress',
        simulator=s,
        index=0,
        info=True,
        see=True,
        reliability=True,
        show=False
    )
    plot.item_exposure(simulator=s, show=False)
    plot.item_exposure(simulator=s, show=False, par='a')
    plot.item_exposure(simulator=s, show=False, par='b')
    plot.item_exposure(simulator=s, show=False, par='c')
    plot.item_exposure(simulator=s, show=False, par='d')
    plot.item_exposure(simulator=s, show=False, hist=True)

    # close all plots after testing
    close('all')


def test_stats():
    import numpy.random as nprnd
    for _ in range(10):
        items = generate_item_bank(500)
        stats.coef_variation(items)
        stats.coef_correlation(items)
        stats.covariance(items)
        stats.covariance(items, False)
        stats.scatter_matrix(items)

        random_integers = nprnd.randint(30, size=1000)
        stats.bincount(random_integers)


if __name__ == '__main__':
    unittest.main()
